/*
	Copyright 2004, Genos Open Source, S.L. (http://genos.accio.com)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
 * Include the core server components.
 */
#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_core.h"

#include <sys/stat.h>
#include <unistd.h>


static void mod_auth_user_dir_register_hooks (apr_pool_t *p);
static int custom_authenticate_basic_user(request_rec *r);
static const command_rec mod_auth_user_dir_cmds[];


#define MODULE_NAME "auth_user_dir"
//#define DEBUG


typedef struct {
	char *string;
} mod_auth_user_dir_config;

static void *create_mod_auth_user_dir_config(apr_pool_t *p, char *dummy)
{
	// This module's configuration structure.
	mod_auth_user_dir_config *newcfg;

	// Allocate memory from the provided pool.
	newcfg = (mod_auth_user_dir_config *) apr_pcalloc(p, sizeof(mod_auth_user_dir_config));

	// Set the string to a default value.
	newcfg->string = NULL; "/home/webdav/";

	// Return the created configuration struct.
	return (void *) newcfg;
}

module AP_MODULE_DECLARE_DATA auth_user_dir_module =
{
	STANDARD20_MODULE_STUFF,
	create_mod_auth_user_dir_config,
	NULL,
	NULL,
	NULL,
	mod_auth_user_dir_cmds,
	mod_auth_user_dir_register_hooks
};


static const char *set_mod_auth_user_dir_string(cmd_parms *parms, void *mconfig, const char *arg)
{

	mod_auth_user_dir_config *s_cfg=(mod_auth_user_dir_config *)mconfig;

	s_cfg->string = (char *) arg;

	return NULL;
}

static const command_rec mod_auth_user_dir_cmds[] =
{
	AP_INIT_TAKE1(
		"AuthUserDirRoot",
		set_mod_auth_user_dir_string,
		NULL,
		ACCESS_CONF,
		"ModAuthUserDirString (string) The error_log string."
	),
	{NULL}
};


/*
 * This function is a callback and it declares what
 * other functions should be called for request
 * processing and configuration requests. This
 * callback function declares the Handlers for
 * other events.
 */
static void mod_auth_user_dir_register_hooks (apr_pool_t *p)
{
	ap_hook_fixups(custom_authenticate_basic_user, NULL, NULL, APR_HOOK_MIDDLE);
}

void denegar_apache(request_rec *r)
{
	ap_note_basic_auth_failure(r);
}


static int custom_authenticate_basic_user(request_rec *r)
{
	int res;
	const char *sent_pw;
	char *path_auth_user;
	char *filename;

	struct stat tipus_arxiu;


	int i,longitut_path_auth_user,longitut_user;
	int inici_dir,final_dir;


	if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
      return res;

	mod_auth_user_dir_config *s_cfg = ap_get_module_config(r->per_dir_config,
		&auth_user_dir_module);

	path_auth_user=s_cfg->string;


#ifdef DEBUG
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": uri=%s user=%s filename=%s path_auth_user:%s",
						 r->uri,r->user,r->filename,path_auth_user);
#endif

	if (path_auth_user==NULL) {
#ifdef DEBUG
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": path_auth_user empty");
#endif
		return DECLINED;
	}


// /home/egroupware/*/ruta..... *=user to authorize

	//path_auth_user=path
	//filename=path asked to apache

	if (r->filename==NULL) {
#ifdef DEBUG
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": filename empty");
#endif
		return DECLINED;
	}

	longitut_path_auth_user=strlen(path_auth_user);

	if (longitut_path_auth_user==0) {
#ifdef DEBUG
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": path_auth_user emty. ok");
#endif
		return DECLINED;
	}

	if (strncmp(path_auth_user,r->filename,longitut_path_auth_user)) {
#ifdef DEBUG
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": %s not in protected path. ok"
							,r->filename);
#endif
		return DECLINED;
	}

	//Allow inside path_auth_user, but authenticate subdirectories

	//Veiem si es dintre subdirectori, o dintre l'arrel de path_auth_user
	filename=r->filename;
	if (!filename[longitut_path_auth_user]) {
		//end of string. filename is path_auth_user without / at the end (for example: "/home/egroupware");

#ifdef DEBUG
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": %s inside root path. ok",r->filename);
#endif

		return DECLINED;
	}

	if (path_auth_user[longitut_path_auth_user-1]!='/') {
		//error. path_auth_user doesn't end with amb /
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": path_auth_user=%s, must end with /",
							path_auth_user);
	  return DECLINED;
	}


	inici_dir=longitut_path_auth_user;
	for (i=inici_dir; filename[i]; i++) {
		if (filename[i]=='/' ){
			//next / found. Look between two /, to match user authenticated
			final_dir=i;
			longitut_user=strlen(r->user);
			//if two lenghts doesnt't match, error
			if (longitut_user!=final_dir-inici_dir) {
#ifdef DEBUG
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": user length (%d) doesn't match with path length (%d)"
									,longitut_user,final_dir-inici_dir);
#endif
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": user: %s not authorized to %s",
									r->user,filename);

				denegar_apache(r);
				return HTTP_UNAUTHORIZED;

			}
			if (strncmp(r->user,&filename[inici_dir],final_dir-inici_dir)) {
				//Error. doesn't match
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": user: %s not authorized to %s",
									r->user,filename);
				denegar_apache(r);
				return HTTP_UNAUTHORIZED;
			}
#ifdef DEBUG
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": %s check valid. user: %s. ok",
								r->filename,r->user);
#endif
			return DECLINED; //OK. Matches user with directory
		}
	}

	//Check if the file is inside root path
	if (stat(r->filename, &tipus_arxiu)==-1) {
		//error with stat
		return DECLINED;
	}
	if (!S_ISDIR(tipus_arxiu.st_mode)) {
#ifdef DEBUG
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": %s inside root path. ok",r->filename);
#endif
		return DECLINED;
	}


	if (strncmp(r->user,&filename[inici_dir],i-inici_dir)) {
		//Check directory without / at the end
		//Error. doens't match
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": user: %s not authorized to %s"
							,r->user,filename);
		denegar_apache(r);
		return HTTP_UNAUTHORIZED;
	}

#ifdef DEBUG
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, MODULE_NAME ": %s inside root path. ok",r->filename);
#endif
	return DECLINED;


}


